#include <pthread.h> 
#include <sched.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <time.h>

typedef struct mybarrier_st {
	int count;
	pthread_mutex_t mutex;
} mybarrier;

mybarrier barrier;

void barrier_init(mybarrier *bar, int count) {
	bar->count = count;
	pthread_mutex_init(&bar->mutex, NULL);
}


void barrier_wait(mybarrier *bar, char *id) {
	printf("Thread %s waiting at barrier...\n", id);

	pthread_mutex_lock(&bar->mutex);
	bar->count--;
	pthread_mutex_unlock(&bar->mutex);

	for (;;) {
		pthread_mutex_lock(&bar->mutex);
		if (bar->count == 0) {
			pthread_mutex_unlock(&bar->mutex);
			break;
		}
		else {
			
			pthread_mutex_unlock(&bar->mutex);
			sched_yield();
		}
	}

	printf("Thread %s passed the barrier!\n", id);
}

void *func(void *arg) {
	char *id = (char *)arg;
	int delay = rand() % 5 + 1;
	printf("Thread %s working for %d sec...\n", id, delay);
	sleep(delay);
	printf("Thread %s done working!\n", id);
	barrier_wait(&barrier, id);
	return NULL;
}

int main(int argc, char *argv[]) {
	srand(time(NULL));

	barrier_init(&barrier, 3);

	printf("Thread main started\n");

	pthread_t threadA, threadB;
	pthread_create(&threadA, NULL, func, "A");
	pthread_create(&threadB, NULL, func, "B");

	barrier_wait(&barrier, "main");

	pthread_join(threadA, NULL);
	pthread_join(threadB, NULL);

	printf("Program completed\n");
	return EXIT_SUCCESS;
}
