#include <stdio.h> 
#include <stdlib.h>
#include <errno.h>
#include <pthread.h>
#include <time.h>
#include <windows.h>
#include <sys/time.h>

#define COUNT  1000000

#define THREAD_COUNT 4

int n;


pthread_mutex_t mutex = PTHREAD_MUTEX_INITIALIZER;
pthread_spinlock_t lock;

void *compute1(void *arg) {
	int i;
	for (i = 0; i < COUNT; i++) {
		pthread_mutex_lock(&mutex);
		n++;
		pthread_mutex_unlock(&mutex);
	}

	return NULL;
}

void *compute2(void *arg) {
	int i;
	for (i = 0; i < COUNT; i++) {
		pthread_spin_lock(&lock);
		n++;
		pthread_spin_unlock(&lock);
	}

	return NULL;
}

void error(char *msg) {
	perror(msg);
	exit(EXIT_FAILURE);
}

void start_time(struct timeval *tval_before) {

	gettimeofday(tval_before, NULL);
}

int elapsed_time(struct timeval tval_before, struct timeval tval_after) {

int microseconds = (tval_after.tv_sec - tval_before.tv_sec) * 1000000 + ((int)tval_after.tv_usec - (int)tval_before.tv_usec);

return microseconds;


}

int main(int argc, char **argv) {
	pthread_t t[THREAD_COUNT];
	int i;

	// Mutex
	struct timeval  t1, t2;
	start_time(&t1);
	n = 0;
	for (i = 0; i < THREAD_COUNT; i++) {
		if (pthread_create(&t[i], NULL, compute1, NULL) != 0)
			error("pthread_create");
	}

	for (i = 0; i < THREAD_COUNT; i++) {
		if (pthread_join(t[i], NULL) != 0)
			error("pthread_join");
	}
	start_time(&t2);
	printf("Time1 elapsed: %.6f\n", elapsed_time(t1,t2)/1000000.0);
	pthread_mutex_destroy(&mutex);

	// Spinlock
	struct timeval  t3, t4;
	start_time(&t3);
	pthread_spin_init(&lock, PTHREAD_PROCESS_PRIVATE);

	n = 0;
	for (i = 0; i < THREAD_COUNT; i++) {
		if (pthread_create(&t[i], NULL, compute2, NULL) != 0)
			error("pthread_create");
	}

	for (i = 0; i < THREAD_COUNT; i++) {
		if (pthread_join(t[i], NULL) != 0)
			error("pthread_join");
	}

	start_time(&t4);
	printf("Time2 elapsed: %.6f\n", elapsed_time(t3,t4)/1000000.0);
	pthread_spin_destroy(&lock);

	return EXIT_SUCCESS;
}
